package org.bdware.consistency.plugin.raft.algo;

import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.pbft.ContractCluster;
import org.bdware.consistency.plugin.raft.algo.message.*;
import org.bdware.consistency.plugin.raft.algo.storage.LogEntry;
import org.bdware.consistency.plugin.raft.algo.storage.LogManager;
import org.bdware.consistency.plugin.raft.algo.storage.StateManager;
import org.bdware.consistency.plugin.raft.algo.util.HeartBeatUtil;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.Node;
import org.bdware.sc.consistency.CommitAlgorithm;
import org.bdware.sc.consistency.Committer;
import org.bdware.sc.units.PubKeyNode;
import org.bdware.sc.units.TrustfulExecutorConnection;

import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class RaftAlgorithm implements CommitAlgorithm {
    private static final Logger LOGGER = org.apache.logging.log4j.LogManager.getLogger(RaftAlgorithm.class);
    private ContractCluster connection;
    Committer committer;

    private enum State {LEADER, FOLLOWER, CANDIDATE}

    State state;
    PubKeyNode leader;
    final PubKeyNode self;
    List<PubKeyNode> servers;

    StateManager stateManager;
    // Persistent state on all servers
    long currentTerm;
    PubKeyNode votedFor;
    LogManager logManager;

    // Volatile state on all servers
    long commitIndex;
    long lastApplied;

    // Volatile state on leaders
    Map<String, Long> nextIndex;
    Map<String, Long> matchIndex;


    int grantedVoteCount;

    // electionTimeout
    TimerTask electionTimerTask;
    TimerTask heartbeatTimerTask;
    long lastActivated;

    //config
    RaftConfig config;

    // Locks and Conditions
    private Lock lock = new ReentrantLock();


    public RaftAlgorithm(RaftConfig config, PubKeyNode node, ContractCluster conn, Committer committer) {
        this.config = config;
        self = node;
        connection = conn;
        this.committer = committer;
        state = State.FOLLOWER;
        leader = null;
        servers = connection.getNodes();
        currentTerm = StateManager.loadCurrentTerm();
        votedFor = StateManager.loadVotedFor();
        logManager = new LogManager(config.raftLogDir, config.maxSegmentFileSize);
        commitIndex = 0;
        lastApplied = 0;

        electionTimerTask = new TimerTask() {
            @Override
            public void run() {
                long cur = System.currentTimeMillis();
                if (cur - lastActivated > config.getElectionTimeout()) {
                    startElection();
                }
            }
        };
        heartbeatTimerTask = new TimerTask() {
            @Override
            public void run() {
                lastActivated = System.currentTimeMillis();
                broadcastAppendEntries();
            }
        };
        resetElectionTimer();
    }


    public boolean isLeader() {
        return state == State.LEADER;
    }

    @Override
    public void onMessage(Node sender, byte[] msg) {
        if (((PubKeyNode) sender).pubkey.equals(self.pubkey)) {
            LOGGER.info("IGNORE SELF:" + msg.length);
            return;
        }
        LOGGER.info("RECEIVE:" + msg.length);
        if (!isServerValid((PubKeyNode) sender)) {
            LOGGER.debug("message from an invalid server ({})!", sender.toString());
            return;
        }
        IRaftMessage message = RaftMessageParser.parse(msg);
        LOGGER.info("RECEIVE:" + message.getType());
        lock.lock();
        try {
            IRaftMessage response;
            switch (message.getType()) {
                case RequestVote:
                    response = onRequestVote((RequestVote) message);
                    persist();
                    connection.sendMessage((PubKeyNode) sender, new RaftMessageParser(response).getBytes());
                    break;
                case RequestVoteResponse:
                    onRequestVoteResponse((RequestVoteResponse) message);
                    break;
                case AppendEntries:
                    response = onAppendEntries((AppendEntries) message);
                    persist();
                    connection.sendMessage((PubKeyNode) sender, new RaftMessageParser(response).getBytes());
                    break;
                case AppendEntriesResponse:
                    onAppendEntriesResponse((PubKeyNode) sender, (AppendEntriesResponse) message);
                    break;
                case Unknown:
                    break;
                default:
                    throw new IllegalStateException("Unexpected value: " + message.getType());
            }
        } finally {
            lock.unlock();
        }
    }

    // in lock
    private RequestVoteResponse onRequestVote(RequestVote request) {
        if (request.getTerm() < currentTerm) {
            LOGGER.debug("Rejected RequestVote message in term {}" +
                    ", because the message is stale. My term is {}.", request.getTerm(), currentTerm);
            return new RequestVoteResponse(currentTerm, false);
        }
        if (request.getTerm() == currentTerm) {
            if (votedFor != null && !votedFor.pubkey.equals(request.getCandidate().pubkey)) {
                LOGGER.debug("Rejected RequestVote message in term {}" +
                        ", because I have voted for another server. ", request.getTerm());
                return new RequestVoteResponse(currentTerm, false);
            }
        } else {
            currentTerm = request.getTerm();
            leader = null;
            votedFor = null;
            convertToFollower();
        }
        if (!isUpToDateThanMyLog(request.getLastLogTerm(), request.getLastLogIndex())) {
            LOGGER.debug("Rejected RequestVote message, because my log is more up-to-date. ");
            return new RequestVoteResponse(currentTerm, false);
        }
        // (request.getTerm() == currentTerm && votedFor == null) || request.getTerm() > currentTerm
        // || request server's log is more up-to-date than mine
        votedFor = request.getCandidate();
        LOGGER.info("Granted RequestVote message from server (pubKey={}) in term {}",
                request.getCandidate().pubkey, currentTerm);

        return new RequestVoteResponse(currentTerm, true);
    }

    // in lock
    private void onRequestVoteResponse(RequestVoteResponse response) {
        if (response.getTerm() > currentTerm) {
            currentTerm = response.getTerm();
            leader = null;
            votedFor = null;
            convertToFollower();
        } else if (state == State.CANDIDATE && response.getTerm() == currentTerm && response.isGranted()) {
            ++grantedVoteCount;
            if (grantedVoteCount > servers.size() / 2) {
                convertToLeader();
                LOGGER.info("Get votes from a majority. Convert to leader.");
            }
        }
    }

    // in lock
    private AppendEntriesResponse onAppendEntries(AppendEntries request) {
        LogEntry lastEntry = logManager.getLastEntry();
        if (request.getTerm() < currentTerm) {
            LOGGER.debug("Rejected AppendEntries message in term {}" +
                    ", because the message is stale. My term is {}.", request.getTerm(), currentTerm);
            return new AppendEntriesResponse(currentTerm, lastEntry.getIndex(), false, self);
        }
        if (request.getTerm() > currentTerm) {
            currentTerm = request.getTerm();
            leader = null;
            votedFor = null;
            convertToFollower();
        }
        if (leader == null) {
            leader = request.getLeader();
            LOGGER.info("new leader={}", leader.pubkey);
        }
        if (!leader.pubkey.equals(request.getLeader().pubkey)) {
            LOGGER.warn("Another peer={} declares that it is the leader " +
                            "at term={} which was occupied by leader={}",
                    request.getLeader().pubkey, request.getTerm(), leader.pubkey);
            return new AppendEntriesResponse(currentTerm, lastEntry.getIndex(), false, self);
        }
        resetElectionTimer();
        lastActivated = System.currentTimeMillis();
        LogEntry entry = logManager.getEntry(request.getPrevLogIndex());
        if (entry == null) {
            return new AppendEntriesResponse(currentTerm, lastEntry.getIndex(), false, self);
        }
        if (entry.getTerm() != request.getTerm() || logManager.getLastEntryIndex() > request.getPrevLogIndex()) {
            logManager.deleteEntriesStartAt(request.getPrevLogIndex());
            lastEntry = logManager.getLastEntry();
        }
        List<LogEntry> newEntries = request.getEntries();
        if (newEntries.size() == 0) {
            return new AppendEntriesResponse(currentTerm, lastEntry.getIndex(), true, self);
        }
        logManager.append(newEntries);
        lastEntry = logManager.getLastEntry();
        if (request.getLeaderCommit() > commitIndex) {
            commitIndex = Math.min(request.getLeaderCommit(), newEntries.get(newEntries.size() - 1).getIndex());
        }
        LOGGER.info("RECEIVE APPENDENTRY: commitIndex=" + commitIndex + " lastApplied:" + lastApplied);
        while (commitIndex > lastApplied) {
            applyToStateMachine(lastApplied + 1);
            lastApplied++;
        }
        return new AppendEntriesResponse(currentTerm, lastEntry.getIndex(), true, self);
    }

    // in lock
    private void onAppendEntriesResponse(PubKeyNode sender, AppendEntriesResponse response) {
        if (response.getTerm() > currentTerm) {
            currentTerm = response.getTerm();
            leader = null;
            votedFor = null;
            convertToFollower();
        } else if (state == State.LEADER && response.getTerm() == currentTerm) {
            if (response.isSuccess()) {
                if (matchIndex.containsKey(sender.pubkey) && matchIndex.get(sender.pubkey) > response.getLastLogIndex()) {
                    LOGGER.warn("node={} matchIndex decrease! prev matchIndex={}, lastLogIndex={}",
                            sender.pubkey, matchIndex.get(sender.pubkey), response.getLastLogIndex());
                    throw new RuntimeException("matchIndex error");
                }
                matchIndex.put(sender.pubkey, response.getLastLogIndex());
            }
            nextIndex.put(sender.pubkey, response.getLastLogIndex() + 1);
        }
    }

    // todo
    private void applyToStateMachine(long logIndex) {
        LogEntry entry = logManager.getEntry(logIndex);
        LOGGER.info("Commit: requestID = " + entry.getCommand().getRequestID());
        committer.onCommit(entry.getCommand());
    }

    public void insertLogEntry(ContractRequest req) {
        LogEntry entry = new LogEntry();
        entry.setCommand(req);
        entry.setIndex(logManager.getLastEntryIndex() + 1);
        entry.setTerm(currentTerm);
        logManager.append(entry);
        broadcastAppendEntries();
    }

    private void broadcastAppendEntries() {
        lock.lock();
        try {
            LogEntry lastEntry = logManager.getLastEntry();
            long lastLogIndex = lastEntry.getIndex();
            nextIndex.forEach((pubKey, index) -> {
                PubKeyNode node = new PubKeyNode();
                node.pubkey = pubKey;
                if (lastLogIndex >= index) {
                    sendAppendEntries(node, index);
                } else {
                    AppendEntries req = new AppendEntries(currentTerm, self, lastEntry.getIndex(), lastEntry.getTerm(), commitIndex, null);
                    connection.sendMessage(node, new RaftMessageParser(req).getBytes());
                }
            });
            List<Long> indexList = new ArrayList<>(matchIndex.values());
            Collections.sort(indexList);

            long majorityIndex = indexList.get((indexList.size() - 1) / 2);
            LogEntry entry = logManager.getEntry(majorityIndex);
            if (majorityIndex > commitIndex && entry.getTerm() == currentTerm) {
                commitIndex = majorityIndex;
            }
            while (commitIndex > lastApplied) {
                applyToStateMachine(lastApplied + 1);
                lastApplied++;
            }
        } finally {
            lock.unlock();
        }
    }

    private void sendAppendEntries(PubKeyNode follower, long logIndex) {
        LogEntry prevLog = logManager.getEntry(logIndex - 1);
        List<LogEntry> entries = logManager.getEntriesStartAt(logIndex);
        AppendEntries request = new AppendEntries(currentTerm, leader, prevLog.getIndex(), prevLog.getTerm(), commitIndex, entries);
        connection.sendMessage(follower, new RaftMessageParser(request).getBytes());
    }

    private void persist() {
        StateManager.saveCurrentTerm(currentTerm);
        StateManager.saveVotedFor(votedFor);
    }

    public void convertToFollower() {
        if (state == State.FOLLOWER) {
            return;
        }
        state = State.FOLLOWER;
        stopHeartbeat();
    }

    // in lock
    public void convertToLeader() {
        state = State.LEADER;
        leader = self;
        nextIndex = new HashMap<>();
        matchIndex = new HashMap<>();
        startHeartbeat();
        LogEntry lastEntry = logManager.getLastEntry();
        for (PubKeyNode server : servers) {
            if (server != null) {
                if (lastEntry == null)
                    nextIndex.put(server.pubkey, 1L);
                else
                    nextIndex.put(server.pubkey, lastEntry.getIndex() + 1);
                matchIndex.put(server.pubkey, 0L);
                //  AppendEntries req = new AppendEntries(currentTerm, self, lastEntry.getIndex(), lastEntry.getTerm(), commitIndex, null);
                //  connection.sendMessage(server,new RaftMessageParser(req).getBytes());
            }
        }
    }

    private void stopHeartbeat() {
        HeartBeatUtil.getInstance().cancel(heartbeatTimerTask);
    }

    private void startHeartbeat() {
        HeartBeatUtil.getInstance().schedule(heartbeatTimerTask, 0, config.heartbeatTimeout);
    }

    private void startElection() {
        RequestVote request;
        lock.lock();
        try {
            ++currentTerm;
            state = State.CANDIDATE;
            votedFor = self;
            grantedVoteCount = 1;
            LogEntry lastLog = logManager.getLastEntry();
            request = new RequestVote(currentTerm, self, lastLog.getIndex(), lastLog.getTerm());
        } finally {
            lock.unlock();
        }
        connection.broadcast(request.getBytes());
    }

    private void resetElectionTimer() {
        HeartBeatUtil.getInstance().cancel(electionTimerTask);
        HeartBeatUtil.getInstance().schedule(electionTimerTask, config.getElectionTimeout(), 1000);
    }


    private boolean isServerValid(PubKeyNode server) {
        if (server == null) return false;
        List<PubKeyNode> nodes = connection.getNodes();
        for (PubKeyNode node : nodes) {
            if (node.pubkey.equals(server.pubkey)) {
                return true;
            }
        }
        return false;
    }

    private boolean isUpToDateThanMyLog(long lastLogTerm, long lastLogIndex) {
        LogEntry myLastLog = logManager.getLastEntry();
        long myLastLogIndex = myLastLog.getIndex();
        long myLastLogTerm = myLastLog.getTerm();
        return lastLogTerm > myLastLogTerm
                || (lastLogTerm == myLastLogTerm && lastLogIndex >= myLastLogIndex);
    }

    @Override
    public void setCommitter(Committer c) {
        committer = c;
    }

    @Override
    public void setConnection(TrustfulExecutorConnection c) {
        connection = (ContractCluster) c;
    }

    public PubKeyNode getLeader() {
        return leader;
    }
}
