package org.bdware.consistency.plugin.raft.algo.storage;

import org.apache.commons.io.FileUtils;
import org.apache.logging.log4j.Logger;
import org.bdware.consistency.plugin.raft.algo.util.FileUtil;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.List;

public class Segment {
    private static final Logger LOGGER = org.apache.logging.log4j.LogManager.getLogger(Segment.class);
    private long startIndex;
    private long endIndex;
    private long fileSize;
    private String dir;
    private String fileName;
    private RandomAccessFile randomAccessFile;
    private List<Record> records;

    public void close() {
        FileUtil.closeFile(randomAccessFile);
    }

    public void setMode(String mode) {
        FileUtil.closeFile(randomAccessFile);
        randomAccessFile = FileUtil.openFile(dir, fileName, mode);
    }

    public static class Record {
        public long offset;
        public LogEntry entry;

        public Record(long offset, LogEntry entry) {
            this.offset = offset;
            this.entry = entry;
        }
    }

    // Create new segment file
    public Segment(String dir, String name) {
        this.dir = dir;
        fileName = name;
        records = new ArrayList<>();
    }

    public Segment(String dir) {
        this.dir = dir;
        records = new ArrayList<>();
    }

    public LogEntry getEntry(long index) {
        if (records.size() == 0 || index < startIndex || index > endIndex) {
            return null;
        }
        int indexInList = (int) (index - startIndex);
        return records.get(indexInList).entry;
    }

    public List<LogEntry> getEntriesNotLessThan(long index) {
        List<LogEntry> entries = new ArrayList<>();
        if (records.size() == 0 || index > endIndex) {
            return entries;
        }
        int indexInList = Math.max(0, (int) (index - startIndex));
        for (int i = indexInList; i < records.size(); i++) {
            entries.add(records.get(i).entry);
        }
        return entries;
    }

    public void deleteEntriesStartAt(long index) {
        if (records.size() == 0 || index > endIndex) {
            return;
        }
        int indexInList = (int) (index - startIndex);
        if (indexInList <= 0) {
            records.clear();
            FileUtil.closeFile(randomAccessFile);
            FileUtil.deleteFile(dir, fileName);
            fileSize = 0;
            return;
        }
        endIndex = index - 1;
        fileSize = records.get(indexInList).offset;
        records.removeAll(records.subList(indexInList, records.size()));
        FileUtil.truncate(randomAccessFile, fileSize);
        FileUtil.closeFile(randomAccessFile);
        String oldFullFileName = dir + File.separator + fileName;
        fileName = String.format("open-%016x", startIndex);
        String newFullFileName = dir + File.separator + fileName;
        FileUtil.moveFile(oldFullFileName, newFullFileName);
        randomAccessFile = FileUtil.openFile(dir, fileName, "rw");
    }

    public List<LogEntry> getEntries() {
        return getEntriesNotLessThan(startIndex);
    }

    public LogEntry getLastEntry() {
        return getEntry(endIndex);
    }

    public void append(LogEntry entry) {
        if (randomAccessFile == null) {
            randomAccessFile = FileUtil.createFile(dir, fileName, "rw");
        }
        if (records.isEmpty()) {
            startIndex = entry.getIndex();
        }
        try {
            records.add(new Record(randomAccessFile.getFilePointer(), entry));
            FileUtil.writeLogEntryToFile(randomAccessFile, entry);
            endIndex = entry.getIndex();
            fileSize = randomAccessFile.length();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static Segment loadSegment(String dir, String fileName) {
        Segment segment = new Segment(dir);
        try {
            String[] splitArray = fileName.split("-");
            if (splitArray.length != 2) {
                LOGGER.warn("segment filename[{}] is not valid", fileName);
                throw new RuntimeException("segment filename[{}] is not valid" + fileName);
            }
            segment.fileName = fileName;
            if (splitArray[0].equals("open")) {
                segment.randomAccessFile = FileUtil.openFile(dir, fileName, "rw");
            } else {
                segment.randomAccessFile = FileUtil.openFile(dir, fileName, "r");
            }
            segment.fileSize = segment.randomAccessFile.length(); // ioException
            long offset = 0;
            while (offset < segment.fileSize) {
                LogEntry entry = FileUtil.readLogEntryFromFile(segment.randomAccessFile);
                if (entry == null) {
                    throw new RuntimeException("read segment log failed");
                }
                Record record = new Record(offset, entry);
                segment.records.add(record);
                offset = segment.randomAccessFile.getFilePointer(); // ioException
            }
            int entrySize = segment.records.size();
            if (entrySize > 0) {
                segment.startIndex = segment.getRecords().get(0).entry.getIndex();
                segment.endIndex = segment.getRecords().get(entrySize - 1).entry.getIndex();
            }
            return segment;
        } catch (IOException ioException) {
            LOGGER.warn("loadSegment exception:", ioException);
            throw new RuntimeException("loadSegment error" + ioException);
        }
    }

    public long getStartIndex() {
        return startIndex;
    }

    public long getEndIndex() {
        return endIndex;
    }

    public long getFileSize() {
        return fileSize;
    }

    public String getFileName() {
        return fileName;
    }

    public RandomAccessFile getRandomAccessFile() {
        return randomAccessFile;
    }

    public List<Record> getRecords() {
        return records;
    }

}
